﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using NTERA.Console.RenderItem;

namespace NTERA.Console
{
	public class ConsoleRenderer
	{
		public int LineHeight = 16;

		public List<List<IRenderItem>> Items { get; protected set; } = new List<List<IRenderItem>>();

		public List<Rectangle> UpdateRegions = new List<Rectangle>();
		
		public int offset;

		public int LastLineCount;

		public delegate void AddedItemEventHandler(IRenderItem item);
		public event AddedItemEventHandler AddedItem;

		public ConsoleRenderer()
		{
			Items.Add(new List<IRenderItem>());
		}

		public void WriteItem(IRenderItem item)
		{
			var list = Items[Items.Count - 1];

			var lastItem = list.LastOrDefault();

			if (lastItem != null)
			{
				if (lastItem is TextRenderItem lastTextItem && item is TextRenderItem newTextItem)
				{
					if (lastTextItem.TextBrush.Color == newTextItem.TextBrush.Color)
					{
						lastTextItem.Text += newTextItem.Text;
						
						AddedItem?.Invoke(item);
						return;
					}
				}
			}

			Items[Items.Count - 1].Add(item);

			AddedItem?.Invoke(item);
		}

		public void PrintItem(IRenderItem item)
		{
			Items[Items.Count - 1].Add(item);
			Items.Add(new List<IRenderItem>());

			AddedItem?.Invoke(item);
		}

		public int Render(Graphics graphics, Rectangle clientArea, Rectangle invalidateArea, Point mousePointer)
		{
			int lastItem = Items.Count - offset;

			int screenLineCount = (int)Math.Ceiling(clientArea.Height / (float)LineHeight);

			int firstItem = Math.Max(0, lastItem - screenLineCount);
			
			for (int i = firstItem; i < lastItem; i++)
			{
				int x = 0;

				int y = clientArea.Height - (lastItem - i) * LineHeight;

				var itemArea = new Rectangle(x, y, clientArea.Width - x, LineHeight);

				if (!invalidateArea.IntersectsWith(itemArea))
					continue;

				itemArea.Intersect(invalidateArea);
				itemArea.Y -= LineHeight - itemArea.Height;
				itemArea.Height = LineHeight;

				foreach (var renderItem in Items[i].ToArray())
				{
					int newX = renderItem.Render(graphics, itemArea, invalidateArea, mousePointer);

					itemArea.Width = newX - x;
					itemArea.X = newX;

					renderItem.Region = itemArea;

					x = newX;
				}
			}

			LastLineCount = screenLineCount;
			return screenLineCount;
		}


		protected Rectangle? lastUpdateRange = null;

		public void MouseHoverEvent(Point mousePoint, Control control)
		{
			foreach (var item in Items.SelectMany(x => x))
			{
				if (!item.InvalidateOnMouseStateChange || !item.Region.Contains(mousePoint))
					continue;

				if (item.Region != lastUpdateRange)
				{
					control.Invalidate(item.Region, true);
					lastUpdateRange = item.Region;
				}
							
				return;
			}

			if (lastUpdateRange != null)
			{
				control.Invalidate(Rectangle.Round(lastUpdateRange.Value), true);
				lastUpdateRange = null;
			}
		}
	}
}