﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using NTERA.Core.Interop;

namespace NTERA.Console.RenderItem
{
	public class ImageRenderItem : BaseRenderItem
	{
		public Bitmap Image { get; set; }

		public DisplayLineAlignment Alignment;

		public ImageRenderItem(Bitmap bitmap, Rectangle? cropRectangle = null, DisplayLineAlignment alignment = DisplayLineAlignment.LEFT)
		{
			if (cropRectangle != null)
			{
				Image = new Bitmap(bitmap.Width, cropRectangle.Value.Height, PixelFormat.Format32bppArgb);

				using (Graphics g = Graphics.FromImage(Image))
				{
					g.DrawImage(bitmap, new Rectangle(0, 0, bitmap.Width, cropRectangle.Value.Height), cropRectangle.Value, GraphicsUnit.Pixel);
				}
			}
			else
			{
				Image = bitmap;
			}
			
			Alignment = alignment;
		}

		public override int Render(Graphics graphics, Rectangle renderArea, Rectangle invalidatedArea, Point mousePointer)
		{
			int x;

			switch (Alignment)
			{
				default:
				case DisplayLineAlignment.LEFT:
					x = renderArea.X;
					break;
				case DisplayLineAlignment.CENTER:
					x = (renderArea.Width - Image.Width) / 2;
					break;
				case DisplayLineAlignment.RIGHT:
					x = renderArea.Width - Image.Width;
					break;
			}

			float ratio = Image.Width / (float)Image.Height;
			int width = (int)(ratio * renderArea.Height);

			graphics.DrawImage(Image, new Rectangle(x, renderArea.Y, width, renderArea.Height));

			return x + width;
		}

		public static List<ImageRenderItem> CreateFromLargeBitmap(Bitmap original, int lineHeight)
		{
			List<ImageRenderItem> items = new List<ImageRenderItem>();

			int total = (int)Math.Ceiling(original.Height / (float)lineHeight);

			Rectangle destRect = new Rectangle(0, 0, original.Width, lineHeight);

			for (int i = 0; i < total; i++)
			{
				//Bitmap temp = new Bitmap(original.Width, lineHeight, PixelFormat.Format32bppArgb);

				//using (Graphics g = Graphics.FromImage(temp))
				//{
				//	g.DrawImage(original, destRect, new Rectangle(0, i * lineHeight, original.Width, lineHeight), GraphicsUnit.Pixel);
				//}

				items.Add(new ImageRenderItem(original, destRect));
			}

			return items;
		}

		public override void Dispose()
		{
			Image?.Dispose();
		}
	}
}